package mfsm.jssp;

import java.io.BufferedWriter;
import java.io.Closeable;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Objects;

/** Create a problem that logs all improvements */
public class Log implements Closeable {

  /** the destination writer */
  private final BufferedWriter m_dest;

  /** the consumed FEs */
  private long m_fes;

  /** the best result */
  private int m_best;

  /** the maximum FEs */
  private final long m_maxFEs;

  /** the best record */
  private final int[][] m_rec;
  /** the actual problem */
  private final Problem m_wrapped;

  /**
   * create
   *
   * @param wrap
   *          the actual problem to wrap
   * @param logFile
   *          the log file
   */
  public Log(final Problem wrap, final Path logFile) {
    this(wrap, logFile, 1_000_000);
  }

  /**
   * create
   *
   * @param wrap
   *          the actual problem to wrap
   * @param logFile
   *          the log file
   * @param maxFEs
   *          the maximum FEs
   */
  public Log(final Problem wrap, final Path logFile, final long maxFEs) {
    super();

    try {
      this.m_dest = Files.newBufferedWriter(logFile);
      this.m_dest.write("FEs\tBestF"); //$NON-NLS-1$
      this.m_dest.newLine();
    } catch (final IOException ioe) {
      throw new IllegalStateException(ioe);
    }
    this.m_fes = 0L;
    this.m_best = Integer.MAX_VALUE;
    this.m_wrapped = Objects.requireNonNull(wrap);
    this.m_rec = wrap.newEmptySolution();
    this.m_maxFEs = maxFEs;
  }

  /** {@inheritDoc} */
  @Override
  public final void close() {
    try {
      try {
        this.m_dest.write(Long.toString(this.m_fes) + '\t'
            + Integer.toString(this.m_best));
        this.m_dest.newLine();
        this.m_dest.write("# end"); //$NON-NLS-1$
        this.m_dest.newLine();
        this.m_dest.write('#');
        this.m_dest.newLine();
        this.m_dest.write('#');
        this.m_dest.newLine();
        this.m_wrapped.print(this.m_rec, this.m_dest);
        this.m_dest.flush();
      } finally {
        this.m_dest.close();
      }
    } catch (final IOException ioe) {
      throw new IllegalStateException(ioe);
    }
  }

  /**
   * Evaluate the solution
   *
   * @param solution
   *          the solution
   * @return the result
   */
  public final int evaluate(final int[][] solution) {
    final int res = this.m_wrapped.evaluate(solution);
    ++this.m_fes;
    if (res < this.m_best) {
      this.m_best = res;
      Log.__copyFromTo(solution, this.m_rec);
      try {
        this.m_dest.write(Long.toString(this.m_fes) + '\t'
            + Integer.toString(this.m_best));
        this.m_dest.newLine();
      } catch (final IOException ioe) {
        throw new IllegalStateException(ioe);
      }
    }
    return res;
  }

  /**
   * Should we terminate?
   *
   * @return {@code true} if we should quit, {@code false} to continue
   */
  public boolean shouldTerminate() {
    return (this.m_fes >= this.m_maxFEs);
  }

  /**
   * Get the best result
   *
   * @return the best result
   */
  public final int[][] getBestSolution() {
    return this.m_rec;
  }

  /**
   * Get the best quality
   *
   * @return the best quality
   */
  public final int getBestQuality() {
    return this.m_best;
  }

  /**
   * Copy one two-dimensional integer array to another
   *
   * @param from
   *          the source array
   * @param to
   *          the destination array
   */
  private static final void __copyFromTo(final int[][] from,
      final int[][] to) {
    int i = -1;
    for (final int[] src : from) {
      System.arraycopy(src, 0, to[++i], 0, src.length);
    }
  }
}
