package mfsm.jssp;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

/**
 * <p>
 * An instance of the Job Shop Scheduling Problem (JSSP). This class
 * provides the instances taken from
 * http://people.brunel.ac.uk/~mastjjb/jeb/orlib/jobshopinfo.html plus one
 * small "demo" instance.
 * </p>
 * <p>
 * The problem instance holds all the data that describes one concrete
 * JSSP. A JSSP consists of {@link #machines m} machines and {@link #jobs
 * n} that need to be executed on the machines. Each job consists of also
 * exactly {@code m} sub-jobs: it will utilize each machine. The order in
 * which the sub-jobs must be executed is defined, as well as the time they
 * need. The goal is to assign jobs to machines such in a way that the
 * makespan, i.e., the total time needed to complete all jobs, is
 * minimized.
 * </p>
 */
public class Problem {

  /** the number of machines */
  public final int machines;

  /** the number of jobs */
  public final int jobs;

  /** for each job, the sequence of machines and times */
  public final int[][] job_machine_times;

  /** the instance name */
  public final String id;

  /**
   * load the specified instance
   * 
   * @param instance
   *          the name of the instance
   */
  @SuppressWarnings("null")
  public Problem(final String instance) {
    super();

    int njobs = -1;
    int nmachines = -1;
    int[][] data = null;
    int dataIndex = 0;

    // load the complete jobshop1.txt file until we find the target
    // instance
    // this is inefficient, but well
    try (
        final InputStream is = Problem.class.getResourceAsStream(//
            (instance.equalsIgnoreCase("demo") ? //$NON-NLS-1$
                "demo.txt" : //$NON-NLS-1$
                "jobshop1.txt")); //$NON-NLS-1$
        final InputStreamReader isr = new InputStreamReader(is);
        final BufferedReader br = new BufferedReader(isr)) {

      // this string indicates that we have the instance
      final String seek = "instance " + instance;//$NON-NLS-1$
      int found = 0;
      String line = null;

      loop: while ((line = br.readLine()) != null) {
        line = line.trim();
        if (line.length() <= 0) {
          continue; // empty line: ignore
        }
        switch (found) {
          case 0: {
            if (seek.equalsIgnoreCase(line)) {
              found = 1; // found instance begin
            }
            break;
          }

          case 3: {
            // found line with the instance size, split at white space
            final String[] s = line.split("\\s+"); //$NON-NLS-1$
            njobs = Integer.parseInt(s[0]);
            nmachines = Integer.parseInt(s[1]);
            if ((njobs <= 0) || (nmachines <= 0)) {
              throw new IllegalArgumentException(//
                  "Invalid instance size: " + //$NON-NLS-1$
                      njobs + ", " + nmachines);//$NON-NLS-1$
            }
            data = new int[njobs][2 * nmachines];
          }

          //$FALL-THROUGH$
          case 1: // ++++
          case 2: { // instance description
            found++; // can ignore instance description
            break; // and ++++ even more so
          }

          case 4: { // instance data: split at white space
            final String[] s = line.split("\\s+"); //$NON-NLS-1$
            if (s.length != data[dataIndex].length) {
              throw new IllegalArgumentException(
                  "Wrong length of job line " + //$NON-NLS-1$
                      (dataIndex + 1));
            }
            for (int i = s.length; (--i) >= 0;) {
              if ((data[dataIndex][i] = Integer.parseInt(s[i])) < 0) {
                throw new IllegalArgumentException(
                    "Illegal machine index/machine time in job line "//$NON-NLS-1$
                        + (dataIndex + 1));
              }
            }
            if ((++dataIndex) >= njobs) {
              break loop;
            }
            break;
          }

          default: {
            throw new IllegalStateException(//
                "Premature end of file.");//$NON-NLS-1$
          }
        }
      }

    } catch (IOException error) {
      throw new IllegalStateException(error);
    }

    // did we find the instance?
    if ((njobs <= 0) || (nmachines <= 0) || (data == null)
        || (dataIndex != data.length)) {
      throw new IllegalArgumentException(//
          "Instance '" + instance + //$NON-NLS-1$
              "' not found."); //$NON-NLS-1$
    }

    // store instance data
    this.jobs = njobs;
    this.machines = nmachines;
    this.job_machine_times = data;
    this.id = instance;
  }

  /**
   * Take a solution, which is basically a GANTT chart, and compute the
   * makespan. Here, a candidate solution is a set of integer arrays, one
   * array per machine. The array for machine {@code i} denotes the jobs
   * that are done at the machine: It is a sequence of
   * {@code JobID1, StartTime1, EndTime1, JobID2, StartTime2, EndTime2, ...}.
   * 
   * @param solution
   *          the candidate solution structure
   * @return the makespan, i.e., the largest end time of any job or machine
   */
  @SuppressWarnings("static-method")
  public final int evaluate(final int[][] solution) {
    int makespan = 0;
    for (int[] machine : solution) {
      int end = machine[machine.length - 1];
      if (end > makespan) {
        makespan = end;
      }
    }
    return makespan;
  }

  /**
   * Create a new, empty candidate solution. This record can be filled with
   * data
   * 
   * @return the new record
   */
  public final int[][] newEmptySolution() {
    return new int[this.machines][this.jobs * 3];
  }

  /**
   * Print a solution as raw data for a Gantt chart.
   * 
   * @param solution
   *          the solution
   * @param dest
   *          the destination
   */
  public void print(final int[][] solution, final Appendable dest) {
    try {
      dest.append("if(!(require(\"plotteR\"))){");//$NON-NLS-1$
      dest.append('\n');
      dest.append("if(!(require(\"devtools\"))){");//$NON-NLS-1$
      dest.append('\n');
      dest.append("install.packages(\"devtools\");");//$NON-NLS-1$
      dest.append('\n');
      dest.append("library(\"devtools\");");//$NON-NLS-1$
      dest.append('\n');
      dest.append("};");//$NON-NLS-1$
      dest.append('\n');
      dest.append("devtools::install_github(\"thomasWeise/plotteR\");");//$NON-NLS-1$
      dest.append('\n');
      dest.append("library(\"plotteR\");");//$NON-NLS-1$
      dest.append('\n');
      dest.append("};");//$NON-NLS-1$
      dest.append('\n');

      dest.append("plot.gantt(list("); //$NON-NLS-1$
      dest.append('\n');

      char next1 = ' ';
      for (final int[] sched : solution) {
        dest.append(next1);
        next1 = ',';
        dest.append("list(");//$NON-NLS-1$
        char next2 = ' ';
        for (int i = 0; i < sched.length;) {
          dest.append(next2);
          next2 = ',';
          dest.append("list(job=");//$NON-NLS-1$
          dest.append(Integer.toString(sched[i++]));
          dest.append("L,start=");//$NON-NLS-1$
          dest.append(Integer.toString(sched[i++]));
          dest.append("L,end=");//$NON-NLS-1$
          dest.append(Integer.toString(sched[i++]));
          dest.append("L)");//$NON-NLS-1$
          dest.append('\n');
        }

        dest.append(')');
        dest.append('\n');
      }

      dest.append("), main=\""); //$NON-NLS-1$
      dest.append(this.id);
      dest.append(", makespan: "); //$NON-NLS-1$
      dest.append(Long.toString(Math.round(this.evaluate(solution))));
      dest.append("\", prefix.job=\"\");"); //$NON-NLS-1$
    } catch (IOException ioe) {
      throw new IllegalStateException(ioe);
    }
  }
}