package mfsm.jssp;

import java.util.Arrays;
import java.util.Random;

/**
 * The encoding of a solution. Here we encode a candidate solution as
 * pseudo-permutation.
 */
public class Representation {

  /** internal buffer variable: the current time at a given machine */
  private final int[] m_machineTime;
  /** internal buffer variable: the step index of the current job */
  private final int[] m_jobState;
  /** internal buffer variable: the time of the current job */
  private final int[] m_jobTime;

  /** for each job, the sequence of machines and times */
  private final int[][] job_machine_times;

  /**
   * create the representation
   * 
   * @param problem
   *          the problem instance
   */
  public Representation(final Problem problem) {
    super();
    this.m_machineTime = new int[problem.machines];
    this.m_jobState = new int[problem.jobs];
    this.m_jobTime = new int[problem.jobs];
    this.job_machine_times = problem.job_machine_times;
  }

  /**
   * Map a point {@code g} from the search space to a candidate solution
   * {@code x} in the solution space.
   * 
   * @param g
   *          the point in the search space
   * @param solution
   *          the solution record, i.e., the Gantt chart
   */
  public void map(final int[] g, final int[][] solution) {
    // setup and initialize the local variables
    final int[] machineTime = this.m_machineTime;
    Arrays.fill(machineTime, 0);
    final int[] jobState = this.m_jobState;
    Arrays.fill(jobState, 0);
    final int[] jobTime = this.m_jobTime;
    Arrays.fill(jobTime, 0);
    final int index = (3 * jobState.length) - 1;

    // fill the schedule with dummy data
    for (final int[] sched : solution) {
      // trick: we use the last slot in each schedule as index
      Arrays.fill(sched, -1);
    }

    // iterate over the jobs in the solution
    for (final int nextJob : g) {
      // get the definition of the steps that we need to take for nextJob
      final int[] jobSteps = this.job_machine_times[nextJob];
      // jobState tells us the index in this list for the next step to do,
      // but since the list contains machine/time pars, we multiply by 2
      final int jobStep = (jobState[nextJob]++) << 1; // get step index
      // so we know the machine where the job needs to go next
      final int machine = jobSteps[jobStep]; // get machine

      // compute start time: it is either the current time we have already
      // spent on the job, or the time when the machine is not occupied -
      // whatever happens later
      final int start = Math.max(machineTime[machine], jobTime[nextJob]);
      // the end time is simply the start time plus the time the job needs
      // to spend on the machine
      final int end = start + jobSteps[jobStep + 1]; // end time
      // this holds for both the machine (it will be ready after end) and
      // the job (it can go to the next station after end)
      jobTime[nextJob] = machineTime[machine] = end; // update state

      // update the schedule with the data we have just computed
      final int[] sched = solution[machine];
      sched[++sched[index]] = nextJob;
      sched[++sched[index]] = start;
      sched[++sched[index]] = end;
    }
  }

  /**
   * Copy the data from a point in the search space from the source to the
   * destination
   * 
   * @param from
   *          the source point
   * @param to
   *          the destination
   */
  @SuppressWarnings("static-method")
  public final void copyFromTo(final int[] from, final int[] to) {
    System.arraycopy(from, 0, to, 0, from.length);
  }

  /**
   * Create a new random point in the search space with default values
   * 
   * @param random
   *          the random number generator
   * @return the result
   */
  public int[] createRandom(final Random random) {
    final int jobs = this.m_jobTime.length;
    final int machines = this.m_machineTime.length;
    final int[] res = new int[jobs * machines];
    int index = -1;

    // first create a feasible string which contains each job index m times
    for (int i = 0; i < jobs; i++) {
      for (int j = 0; j < machines; j++) {
        res[++index] = i;
      }
    }
    // then we randomly shuffle the string: the values don't change, just
    // their positions,
    // i.e. we get a representation for a random schedule
    Tools.shuffle(random, res, 0, res.length);
    return (res);
  }

  /**
   * create a randomly modified copy of the given point in search space
   * 
   * @param random
   *          the random number generator
   * @param src
   *          the source point
   * @param dest
   *          the destination point
   */
  @SuppressWarnings("static-method")
  public final void modify(final Random random, final int[] src,
      final int[] dest) {
    System.arraycopy(src, 0, dest, 0, src.length); // make the copy
    for (;;) {
      // repeatedly select two random locations..
      final int i = random.nextInt(dest.length);
      final int j = random.nextInt(dest.length);
      final int ti = dest[i];
      final int tj = dest[j];
      if (ti != tj) {
        // until we find two with two different values
        // then we swap the values
        dest[i] = tj;
        dest[j] = ti;
        return;
      }
    }
  }

  /**
   * Recombine two points in the search space to form a new one
   * 
   * @param random
   *          the random number generator
   * @param src1
   *          the first source
   * @param src2
   *          the second source
   * @param dest
   *          the destination
   */
  public final void recombine(final Random random, final int[] src1,
      final int[] src2, final int[] dest) {
    final int[] done;
    final int max;

    max = this.m_machineTime.length;
    done = this.m_jobTime;
    Arrays.fill(done, 0); // we need to keep track of completed jobs

    int index = (-1);
    for (final int p1 : src1) {
      final int p2 = src2[++index]; // read p2 from src2

      int choice = p1; // by default, we would pick p1
      if (p1 != p2) { // values at same index differ?
        if (done[p1] >= max) { // job p1 is already completed
          choice = p2; // use p2 instead of p1 (but p2 may also be done)
        } else {
          if ((done[p2] < max) && random.nextBoolean()) {
            choice = p2; // otherwise, use p2 with 50% probability
          } // if the job has not yet been completed
        }
      }
      // it could be that our chosen job is already completed
      while (done[choice] >= max) { // the chosen job is completed
        choice = random.nextInt(done.length); // pick random incomplete one
      }

      ++done[choice]; // got to the next step of the job
      dest[index] = choice; // store the job
    }
  }
}