package mfsm.jssp.algos;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Objects;
import java.util.Random;

import mfsm.jssp.Log;
import mfsm.jssp.Problem;
import mfsm.jssp.Representation;
import mfsm.jssp.Tools;

/** A evolutionary algorithm */
public class EA {

  /**
   * Solve a problem with a simple EA with given parameterization
   * 
   * @param problem
   *          the problem
   * @param representation
   *          the representation
   * @param log
   *          the log
   * @param mu
   *          the mu
   * @param lambda
   *          the lambda
   * @param cr
   *          the crossover rate
   * @return the result
   */
  public static int[][] run(final Problem problem,
      final Representation representation, final Log log, final int mu,
      final int lambda, final double cr) {
    // create local variables
    final Random random = new Random();
    final int[][] x = problem.newEmptySolution();
    final Individual[] population = new Individual[mu + lambda];

    // first generation: fill population with random individuals
    for (int i = population.length; (--i) >= 0;) {
      final int[] g = representation.createRandom(random);
      representation.map(g, x);
      population[i] = new Individual(g, log.evaluate(x));
    }

    for (;;) { // main loop: one iteration = one generation
      // sort the population: mu best individuals at front are selected
      Arrays.sort(population);
      // shuffle mating pool to ensure fairness if lambda<mu
      Tools.shuffle(random, population, 0, mu);
      int p1 = -1; // index to iterate over first parent

      // override the worse lambda solutions with new offsprings
      for (int index = population.length; (--index) >= mu;) {
        if (log.shouldTerminate()) {
          return log.getBestSolution(); // return best solution
        }

        final Individual dest = population[index];
        final Individual parent1 = population[(++p1) % mu]; // parent 1

        if (random.nextDouble() <= cr) { // crossover!
          int p2;
          do { // find a second parent who is different from parent 1
            p2 = random.nextInt(mu);
          } while (p2 == p1);

          representation.recombine(random, parent1.g, population[p2].g,
              dest.g); // perform recombination
        } else { // otherwise create modified copy of first parent
          representation.modify(random, parent1.g, dest.g);
        }

        representation.map(dest.g, x); // map to solution/schedule
        dest.quality = log.evaluate(x); // compute quality
      } // the end of the offspring generation
    } // the end of the main loop
  }

  /**
   * the main method
   * 
   * @param args
   *          the args
   * @throws Throwable
   *           if it must
   */
  public static final void main(final String[] args) throws Throwable {
    final Problem problem = new Problem(
        (args.length > 0) ? args[0] : "la24"); //$NON-NLS-1$

    final int mu = (args.length > 1) ? Integer.parseInt(args[1]) : 512;
    final int lambda = (args.length > 2) ? Integer.parseInt(args[2]) : 512;
    final double cr = (args.length > 3) ? Double.parseDouble(args[3])
        : 0.1;

    final String crs = Double.toString(cr);
    final Path output = Paths.get(".")//$NON-NLS-1$
        .resolve("../output")//$NON-NLS-1$
        .resolve(
            (((((EA.class.getSimpleName() + '_') + mu) + '_') + lambda)
                + '_') + crs.substring(crs.indexOf('.') + 1))//
        .resolve(problem.id);
    Files.createDirectories(output);

    final Path file = output.resolve("run_"//$NON-NLS-1$
        + Files.list(output).count()//
        + ".txt");//$NON-NLS-1$

    final int[][] result;
    try (final Log log = new Log(problem, file)) {
      result = run(problem, new Representation(problem), log, mu, lambda,
          cr);
    }

    problem.print(result, System.out);
  }

  /**
   * the individual record: hold one point in search space and its quality
   */
  private static final class Individual implements Comparable<Individual> {
    /** the point in the search space */
    final int[] g;
    /** the quality */
    int quality;

    /**
     * create the individual record
     * 
     * @param _g
     *          the point in the search space
     * @param _q
     *          the quality
     */
    Individual(final int[] _g, final int _q) {
      super();
      this.g = Objects.requireNonNull(_g);
      this.quality = _q;
    }

    /**
     * compare two individuals: the one with smaller quality is better.
     * 
     * @return -1 if this record is better than {@code o}, 1 if it is
     *         worse, 0 otherwise
     */
    @Override
    public final int compareTo(Individual o) {
      return Integer.compare(this.quality, o.quality);
    }
  }
}
