package mfsm.jssp.algos;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Random;

import mfsm.jssp.Log;
import mfsm.jssp.Problem;
import mfsm.jssp.Representation;

/** A simple stochastic hill climbing algorithm with restarts */
public class HillClimberWithRestarts {

  /**
   * Solve a problem with simple stochastic hill climbing with restarts
   * 
   * @param problem
   *          the problem
   * @param representation
   *          the representation
   * @param log
   *          the log
   * @return the result
   */
  public static int[][] run(final Problem problem,
      final Representation representation, final Log log) {
    // allocate variables
    final Random random = new Random();
    final int[][] xcur = problem.newEmptySolution();

    // begin at a random point in the search space
    int[] gbest = representation.createRandom(random);
    representation.map(gbest, xcur);
    int fbest = log.evaluate(xcur);

    // we need a second point for testing
    final int[] gcur = gbest.clone();

    // initialize counters
    long lastImprovement = 0L;
    long currentStep = 0L;
    long stepsBeforeRestart = 1000L; // restart after 1000 failures

    while (!(log.shouldTerminate())) {
      ++currentStep;// increase step counter

      /// create a modified copy of best point in search space
      representation.modify(random, gbest, gcur);
      // and evaluate it
      representation.map(gcur, xcur);
      int fcur = log.evaluate(xcur);

      if (fcur < fbest) {
        // we found a better solution: move to new point
        fbest = fcur;
        representation.copyFromTo(gcur, gbest);
        lastImprovement = currentStep; // we improved!
      } else { // ignore new point
        if ((currentStep - lastImprovement) >= stepsBeforeRestart) {
          gbest = representation.createRandom(random); // begin new
          fbest = Integer.MAX_VALUE;
          lastImprovement = currentStep; // reset counter
          // increase steps before restart by 12.5%
          stepsBeforeRestart += (stepsBeforeRestart >> 4);
        }
      }
    }

    return log.getBestSolution(); // return best solution
  }

  /**
   * the main method
   * 
   * @param args
   *          the args
   * @throws Throwable
   *           if it must
   */
  public static final void main(final String[] args) throws Throwable {
    final Problem problem = new Problem(
        (args.length > 0) ? args[0] : "la24"); //$NON-NLS-1$

    final Path output = Paths.get(".")//$NON-NLS-1$
        .resolve("../output")//$NON-NLS-1$
        .resolve(HillClimberWithRestarts.class.getSimpleName())//
        .resolve(problem.id);
    Files.createDirectories(output);

    final Path file = output.resolve("run_"//$NON-NLS-1$
        + Files.list(output).count()//
        + ".txt");//$NON-NLS-1$

    final int[][] result;
    try (final Log log = new Log(problem, file)) {
      result = run(problem, new Representation(problem), log);
    }

    problem.print(result, System.out);
  }
}
