package mfsm.jssp.algos;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Random;

import mfsm.jssp.Log;
import mfsm.jssp.Problem;
import mfsm.jssp.Representation;

/**
 * A single sample algorithm. This algorithm creates a single random
 * schedule.
 */
public final class SingleSample {

  /**
   * Generate a single random schedule.
   * 
   * @param problem
   *          the problem instance, i.e., the jobs and machines
   * @param representation
   *          the methods provided by our representation
   * @param log
   *          the log: simple class to collect/print results
   * @return the Gantt chart data found
   */
  public static final int[][] run(final Problem problem,
      final Representation representation, final Log log) {

    final Random random = new Random();
    final int[][] solution = problem.newEmptySolution();

    // create a random point in the search space
    int[] g = representation.createRandom(random);

    // map the point to a solution
    representation.map(g, solution);

    // compute quality: actually calls problem.evaluate and remembers best result
    log.evaluate(solution);

    return log.getBestSolution(); // return best logged solution
  }

  /**
   * the main method
   * 
   * @param args
   *          the args
   * @throws Throwable
   *           if it must
   */
  public static final void main(final String[] args) throws Throwable {
    final Problem problem = new Problem(
        (args.length > 0) ? args[0] : "la24"); //$NON-NLS-1$

    final Path output = Paths.get(".")//$NON-NLS-1$
        .resolve("../output")//$NON-NLS-1$
        .resolve(SingleSample.class.getSimpleName())//
        .resolve(problem.id);
    Files.createDirectories(output);

    final Path file = output.resolve("run_"//$NON-NLS-1$
        + Files.list(output).count()//
        + ".txt");//$NON-NLS-1$

    final int[][] result;
    try (final Log log = new Log(problem, file)) {
      result = run(problem, new Representation(problem), log);
    }

    problem.print(result, System.out);
  }
}
